# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/06/2017"


from silx.gui import qt
from tomwer.core.process.reconstruction.ftseries.params.fastsetupdefineglobals import (
    FastSetupAll,
)
from tomwer.gui.linksqobject import LinkComboBox, LinkCheckBox, LinkLineEdit, LinkGroup
from tomwer.gui.utils.inputwidget import SelectionLineEdit
from tomwer.core.process.reconstruction.ftseries.params.reconsparams import _ReconsParam
from tomwer.core.process.reconstruction.ftseries.params.reconsparams import ReconsParams
from tomwer.core.process.reconstruction.ftseries.params.beamgeo import BeamGeoType
import logging
import enum

logger = logging.getLogger(__name__)


class H5StructsEditor:
    """
    Abstract class which allow the edition of n H5 structures
    """

    def __init__(self, structsManaged=None):
        self.defaultValues = {}
        self.dicVarToWidget = {}
        self.paramToWidget = {}
        self.structsManaged = {}

        self._isLoading = False

        if structsManaged is not None:
            assert type(structsManaged) in (list, tuple)
            for s in structsManaged:
                self.structsManaged[s] = {}
                self.defaultValues[s] = FastSetupAll.getDefaultValues(s)
                self.dicVarToWidget[s] = {}

        self.structs = self.structsManaged.keys()

    def isParamH5Managed(self, structID, paramID):
        return (
            structID in self.structsManaged and paramID in self.structsManaged[structID]
        )

    def linkComboboxWithH5Variable(
        self,
        qcombobox,
        structID,
        h5ParamName,
        fitwithindex=False,
        dic=None,
        filters=None,
        setDefault=True,
    ):
        """

        :param qcombobox: the QCombobox in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param structID: the structID associated with the param ID
        :param fitwithindex: true if we want to store the index in the h5File,
            otherwise store the string
        :param filters: filter to apply on the string to get the requested
            value
        """
        assert type(qcombobox) is qt.QComboBox
        self.dicVarToWidget[structID][h5ParamName] = LinkComboBox(
            qcombobox, fitwithindex, dic, filters, setDefault
        )

    def linkCheckboxWithH5Variable(
        self, qcheckbox, structID, h5ParamName, invert=False
    ):
        """

        :param qcheckbox: the QCheckBox in the interface to edit/display the
            given h5ParamName
        :param structID: the structID associated with the param ID
        :param h5ParamName: the parameter ID in the H5 structure.
        :param invert: true if we will not store the value state == checked but
            state == unchecked
        """
        assert type(qcheckbox) is qt.QCheckBox
        self.dicVarToWidget[structID][h5ParamName] = LinkCheckBox(qcheckbox, invert)

    def LinkLineEditWithH5Variable(
        self, qlineedit, structID, h5ParamName, h5paramtype=str
    ):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param structID: the structID associated with the param ID
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        assert type(qlineedit) is qt.QLineEdit
        self.dicVarToWidget[structID][h5ParamName] = LinkLineEdit(
            qlineedit, h5paramtype
        )

    def LinkSelectionLineEditWithH5Variable(
        self, qlineedit, structID, h5ParamName, h5paramtype=str
    ):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param structID: the structID associated with the param ID
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        assert isinstance(qlineedit, SelectionLineEdit)
        self.dicVarToWidget[structID][h5ParamName] = LinkLineEdit(
            qlineedit, h5paramtype
        )

    def linkGroupWithH5Variable(self, group, structID, h5ParamName, getter, setter):
        self.dicVarToWidget[structID][h5ParamName] = LinkGroup(group, getter, setter)

    def getParamValue(self, structID, paramID):
        if not self.isParamH5Managed(structID, paramID):
            return None
        else:
            return self._getParamValue(self, structID, paramID)

    def _getParamValue(self, structID, h5ParamName):
        """Return the value for the stored h5ParamName

        :param str structID: ID of the h5 group
        :param h5ParamName: the name of the parameter we want to get the value for
        """
        if structID not in self.dicVarToWidget:
            mess = "cannot find structur %s in the managed structure " % structID
            logger.error(mess)
            return None

        if h5ParamName not in self.dicVarToWidget[structID]:
            mess = "cannot find parameter %s in the stored parameters." % h5ParamName
            logger.error(mess)
            return None

        link = self.dicVarToWidget[structID][h5ParamName]
        if type(link) is LinkComboBox:
            return self._getQComboboxValue(link)
        elif type(link) is LinkLineEdit:
            return self._getQLineEditValue(link)
        elif type(link) is LinkCheckBox:
            return self._getQCheckBoxValue(link)
        elif type(link) is LinkGroup:
            return link.getter()
        else:
            mess = (
                "type of the LinkQObject (%s) is not managed by H5StructEditor"
                % type(link)
            )
            raise ValueError(mess)

    def getStructs(self):
        """

        :return: the struct represented by the object under the tuple
                 (structID, structParameters)
        """
        res = {}
        for structID in self.dicVarToWidget:
            parameters = {}
            for param in self.dicVarToWidget[structID]:
                parameters[param] = self._getParamValue(structID, param)
            res[structID] = parameters
        return res

    # redefinition of the load function defined in H5StructEditor
    def load(self, allStructs):
        """
        Set the parameters values to corresponding QObject.

        :param allStructs: is the H5file set of parameters. In this case it
                           include nodes FT, PAGANIN...
        """
        if self._isLoading is True:
            return

        assert isinstance(allStructs, _ReconsParam)
        self._isLoading = True
        self.defaults = None

        if isinstance(allStructs, ReconsParams):
            for structID in self.structs:
                self.structsManaged[structID] = []
                if structID not in allStructs.all_params:
                    error = structID + " is not in the given structure. Can't load."
                    error += "Will load the default structure"
                    logger.error(error)
                    recons_param = self.defaultValues[structID]
                else:
                    recons_param = allStructs._get_parameter_value(structID)
                assert recons_param is None or isinstance(
                    recons_param, (_ReconsParam, _ReconsParam)
                )

                for param in self.dicVarToWidget[structID]:
                    self.structsManaged[structID].append(param)
                    paramValue = recons_param._get_parameter_value(
                        param
                    )  # self._values[param]
                    assert type(paramValue) is not dict
                    self.setParam(structID, param, paramValue=paramValue)
        else:
            for structID in self.structsManaged:
                self.structsManaged[structID] = []
                for param in self.dicVarToWidget[structID]:
                    self.structsManaged[structID].append(param)
                    paramValue = allStructs[param]
                    assert not isinstance(paramValue, _ReconsParam)
                    self.setParam(structID, param, paramValue=paramValue)
        if hasattr(self, "_sliceSlectionWidget"):
            self._sliceSlectionWidget.setEnabled(self._recons_params.do_test_slice)
        self._isLoading = False

    def setParam(self, structID, param, paramValue):
        """Set the parameter value for the gicen H5StructEditor param name

        :param str structID: ID of the h5 group
        :param param: ID of the H5 parameter to set
        :param paramValue: the value to set to the given parameter
        """
        assert structID in self.dicVarToWidget
        assert param in self.dicVarToWidget[structID]

        # setting value
        if type(self.dicVarToWidget[structID][param]) is LinkLineEdit:
            self.dicVarToWidget[structID][param].qobject.setText(str(paramValue))
        elif type(self.dicVarToWidget[structID][param]) is LinkCheckBox:
            if self.dicVarToWidget[structID][param].invert is False:
                state = qt.Qt.Checked if bool(paramValue) else qt.Qt.Unchecked
            else:
                state = qt.Qt.Unchecked if bool(paramValue) else qt.Qt.Checked

            self.dicVarToWidget[structID][param].qobject.setCheckState(state)
        elif type(self.dicVarToWidget[structID][param]) is LinkComboBox:
            self._setQComboboxValue(structID, param, paramValue)
        elif type(self.dicVarToWidget[structID][param]) is LinkGroup:
            if self.dicVarToWidget[structID][param].setter is not None:
                self.dicVarToWidget[structID][param].setter(paramValue)
        else:
            message = "parameter %s no managed" % param
            logger.warning(message)

    # --- Getter and setter of the qobjects ---

    def _getQComboboxValue(self, link):
        """
        Return the actual value saved in the GUI
        """
        assert type(link) is LinkComboBox
        value = (
            link.qobject.currentIndex()
            if link.fitwithindex
            else link.qobject.currentText()
        )
        if link.filters is not None:
            for filt in link.filters:
                value = value.replace(filt, "")
        if link.dicCBtoH5 is not None:
            assert value in link.dicCBtoH5
            return link.dicCBtoH5[value]
        else:
            return value

    def _getValueToSetFromDic(self, structID, param, paramValue):
        if isinstance(paramValue, BeamGeoType):
            paramValue = paramValue.name[0]
        if self.dicVarToWidget[structID][param].dicH5ToCB is not None:
            if paramValue not in self.dicVarToWidget[structID][param].dicH5ToCB:
                error = "'%s' is not a correct value for %s." % (paramValue, param)
                error += " Setting to default"
                logger.warning(error)
                return self.dicVarToWidget[structID][param].dicH5ToCB[
                    self.getDefaultValue(structID, param)
                ]
            else:
                return self.dicVarToWidget[structID][param].dicH5ToCB[paramValue]
        else:
            return paramValue

    def _setQComboboxValue(self, structID, param, paramValue):
        """
        Set the value of the QCombobox for the param param of the struct
        StructID

        :param str structID: ID of the h5 group
        :param param: the H5parameter we want to store for setting the
            reconstruction
        :param paramValue: Value of param
        """
        finalParamValue = self._getValueToSetFromDic(structID, param, paramValue)
        if isinstance(finalParamValue, BeamGeoType):
            finalParamValue = finalParamValue.name[0]
        elif isinstance(finalParamValue, enum.Enum):
            finalParamValue = finalParamValue.name
        index = self.dicVarToWidget[structID][param].qobject.findText(finalParamValue)

        if index == -1:
            info = "(Loading H5StructEditor) :"
            info += "For parameter %s" % param
            info += ", value %s not existing yet." % finalParamValue
            info += "Adding 'unregular value. "
            info += "Can be source of error."
            logger.info(info)
            index = self.dicVarToWidget[structID][param].qobject.count()
            self.dicVarToWidget[structID][param].qobject.addItem(str(finalParamValue))

        self.dicVarToWidget[structID][param].qobject.setCurrentIndex(index)

    def _getQLineEditValue(self, link):
        # bas hack : shouldn't exists but int are stored as float...
        assert type(link) is LinkLineEdit
        if link.h5paramtype is int:
            return link.h5paramtype(float(link.qobject.text()))
        else:
            return link.h5paramtype(link.qobject.text())

    def _getQCheckBoxValue(self, link):
        assert type(link) is LinkCheckBox
        checked = link.qobject.isChecked()
        # all checkbox values are stored as int
        return int(not checked if link.invert else checked)

    def getDefaultValue(self, structID, param):
        """
        :return: the default parameter value or None if not existing
        """
        if structID not in self.defaultValues:
            error = (
                "%s structure is not registred in the default parameters." % structID
            )
            raise ValueError(error)

        if param not in self.defaultValues[structID]:
            error = "%s parameter is not registred in the default parameters." % param
            error += "Shouldn't be managed by H5StructEditor"
            raise ValueError(error)

        return self.defaultValues[structID]._get_parameter_value(param)

    def clear(self):
        self.paramToWidget.clear()
        self.structsManaged.clear()
        self.dicVarToWidget.clear()



from nabu.pipeline.fullfield.chunked_cuda import CudaChunkedPipeline
from nabu.pipeline.fullfield.processconfig import ProcessConfig


def launch_reconstruction(config_file, slice_index):
    proc = ProcessConfig(config_file)
    worker_process = CudaChunkedPipeline(
        proc, sub_region=(None, None, slice_index, slice_index + 1)
    )
    worker_process.process_chunk()



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016-2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/
#
# __authors__ = ["H. Payno", "Orange"]
# __license__ = "MIT"
# __date__ = "03/03/2017"
#
# """
# Unit test from the Orange canvas main instance
#
# """
# import os
# import sys
# import gc
# import re
# import logging
# import pickle
# import pkg_resources
# from AnyQt.QtGui import QFont
# from AnyQt.QtCore import Qt, QDir
# import AnyQt.importhooks
#
# if AnyQt.USED_API == "pyqt5":
#     # Use a backport shim to fake leftover PyQt4 imports
#     AnyQt.importhooks.install_backport_hook("pyqt4")
#
# from silx.gui.utils.testutils import TestCaseQt
#
# from Orange import canvas
# from tomwer.gui.qtapplicationmanager import QApplicationManager
# from orangecanvas.application.canvasmain import CanvasMainWindow
# from Orange.canvas.config import cache_dir
# from Orange.canvas import config
# from orangecanvas.registry.base import WidgetRegistry
# from orangecanvas.registry import qt
# from orangecanvas.registry import set_global_registry
# from orangecanvas.scheme import SchemeNode, SchemeLink
# from orangecanvas.registry import cache
# from orangecanvas import config as canvasconfig
# from unittest import TestCase
#
# try:
#     import graypy
# except:
#     graypy = None
# else:
#     from tomwer.web import config as configweb
# log = logging.getLogger(__name__)
#
#
# class OrangeWorkflowMixIn:
#     """MixIn class for the orange workflows"""
#
#     def processOrangeEvents(self):
#         self.canvas_window.current_document().scheme().signal_manager.process_queued()
#
#     def processOrangeEventsStack(self):
#         node_update_front = (
#             self.canvas_window.current_document()
#             .scheme()
#             .signal_manager.node_update_front()
#         )
#         if node_update_front:
#             self.processOrangeEvents()
#             self.processOrangeEventsStack()
#             self.app.processEvents()
#
#     def init(self):
#         # set default application configuration
#         config_ = config.Config()
#         canvasconfig.set_default(config_)
#         self.app = QApplicationManager()
#         config_.init()
#
#         # File handler should always be at least INFO level so we need
#         # the application root level to be at least at INFO.
#
#         rootlogger = logging.getLogger(canvas.__name__)
#
#         if graypy is not None and configweb.grayport_host is not None:
#             # graypy > 1.1
#             if hasattr(graypy, "GELFUDPHandler"):
#                 cls = graypy.GELFUDPHandler
#             else:
#                 cls = graypy.GELFHandler
#             self.graylogHandler = cls(configweb.grayport_host, configweb.grayport_port)
#             rootlogger.addHandler(self.graylogHandler)
#
#         # Standard output stream handler at the requested level
#         stream_hander = logging.StreamHandler()
#         rootlogger.addHandler(stream_hander)
#
#         log.info("Starting 'Orange Canvas' application.")
#         log.debug("Starting CanvasApplicaiton ")
#         # NOTE: config.init() must be called after the QApplication constructor
#
#         file_handler = logging.FileHandler(
#             filename=os.path.join(config.log_dir(), "canvas.log"), mode="w"
#         )
#
#         rootlogger.addHandler(file_handler)
#
#         # intercept any QFileOpenEvent requests until the main window is
#         # fully initialized.
#         # NOTE: The QApplication must have the executable ($0) and filename
#         # arguments passed in argv otherwise the FileOpen events are
#         # triggered for them (this is done by Cocoa, but QApplicaiton filters
#         # them out if passed in argv)
#
#         open_requests = []
#
#         def onrequest(url):
#             log.info("Received an file open request %s", url)
#             open_requests.append(url)
#
#         self.app.fileOpenRequest.connect(onrequest)
#
#         stylesheet = "orange"
#         stylesheet_string = None
#
#         if stylesheet != "none":
#             if os.path.isfile(stylesheet):
#                 with open(stylesheet, "r") as f:
#                     stylesheet_string = f.read()
#             else:
#                 if not os.path.splitext(stylesheet)[1]:
#                     # no extension
#                     stylesheet = os.path.extsep.join([stylesheet, "qss"])
#
#                 pkg_name = canvas.__name__
#                 resource = "styles/" + stylesheet
#
#                 if pkg_resources.resource_exists(pkg_name, resource):
#                     stylesheet_string = pkg_resources.resource_string(
#                         pkg_name, resource
#                     ).decode()
#
#                     base = pkg_resources.resource_filename(pkg_name, "styles")
#
#                     pattern = re.compile(
#                         r"^\s@([a-zA-Z0-9_]+?)\s*:\s*([a-zA-Z0-9_/]+?);\s*$",
#                         flags=re.MULTILINE,
#                     )
#
#                     matches = pattern.findall(stylesheet_string)
#
#                     for prefix, search_path in matches:
#                         QDir.addSearchPath(prefix, os.path.join(base, search_path))
#                         log.info(
#                             "Adding search path %r for prefix, %r", search_path, prefix
#                         )
#
#                     stylesheet_string = pattern.sub("", stylesheet_string)
#
#                 else:
#                     log.info("%r style sheet not found.", stylesheet)
#
#         # Add the default canvas_icons search path
#         dirpath = os.path.abspath(os.path.dirname(canvas.__file__))
#         QDir.addSearchPath("canvas_icons", os.path.join(dirpath, "icons"))
#         self.canvas_window = CanvasMainWindow()
#         self.canvas_window.setWindowIcon(config.application_icon())
#
#         if stylesheet_string is not None:
#             self.canvas_window.setStyleSheet(stylesheet_string)
#
#         reg_cache = None
#         widget_discovery = qt.QtWidgetDiscovery(cached_descriptions=reg_cache)
#
#         widget_registry = qt.QtWidgetRegistry()
#         widget_discovery = config_.widget_discovery(
#             widget_registry, cached_descriptions=reg_cache)
#
#         try:
#             widget_discovery.run(config.widgets_entry_points(False))
#         except:
#             widget_discovery.run(config.widgets_entry_points())
#
#         set_global_registry(widget_registry)
#         self.canvas_window.set_widget_registry(widget_registry)
#         self.canvas_window.show()
#         self.widget_registry = widget_registry
#
#         # Process events to make sure the self.canvas_window layout has
#         # a chance to activate (the welcome dialog is modal and will
#         # block the event queue, plus we need a chance to receive open file
#         # signals when running without a splash screen)
#         self.app.processEvents()
#         log.info("discovery ended")
#         # Tee stdout and stderr into Output dock
#         log.info("End initialization")
#
#     def addWidget(self, widgetID):
#         """Create an instanciation of the widgetID
#
#         :return SchemeNode: the node created
#         """
#         assert type(widgetID) is str
#         assert self.widget_registry.has_widget(widgetID)
#         widget_desc = self.widget_registry.widget(widgetID)
#         myItem = SchemeNode(widget_desc)
#         self.canvas_window.current_document().scheme().add_node(myItem)
#         return myItem
#
#     def removeNode(self, widget):
#         self.canvas_window.current_document().scheme().remove_node(widget)
#
#     def link(self, schemeNode1, outNode1, schemeNode2, inputNode2):
#         """Create a link between the node on the given input and output
#
#         :param SchemeNode schemeNode1: the node emitting the output signal
#         :param outNode1: the output chanel to be linked
#         :param SchemeNode schemeNode2: the node receiving the input signal
#         :param inputNode2: the input chanel to be linked
#         :return: the link created between the two nodes
#         """
#         assert type(schemeNode1) is SchemeNode
#         assert type(schemeNode2) is SchemeNode
#         assert type(outNode1) is str
#         assert type(inputNode2) is str
#         link = SchemeLink(schemeNode1, outNode1, schemeNode2, inputNode2)
#         self.canvas_window.current_document().scheme().add_link(link)
#         return link
#
#     def getWidgetForNode(self, node):
#         """
#
#         :param SchemeNode node: the node for which we want the corresponding
#             widget
#         :return: the widget instanciated for the given node
#         """
#         assert type(node) is SchemeNode
#         return self.canvas_window.current_document().scheme().widget_for_node(node)
#
#     def tearDown(self):
#         signal_manager = self.canvas_window.current_document().scheme().signal_manager
#         self.canvas_window.setAttribute(qt.Qt.WA_DeleteOnClose)
#         self.canvas_window.closeEvent = self.closeEventMonkeyPatch
#         self.canvas_window.close()
#         self.canvas_window = None
#         self.widget_registry = None
#
#         self.app.removePostedEvents(signal_manager)
#         try:
#             del self.canvas_window
#         except AttributeError:
#             pass
#         gc.collect()
#         self.app.flush()
#         self.app.quit()
#
#         self.app = None
#
#     @classmethod
#     def closeEventMonkeyPatch(cls, event):
#         """monkey patch clsoe event to skip modification saving request..."""
#         event.ignore()
#
#
# class OrangeWorflowTest(OrangeWorkflowMixIn, TestCaseQt):
#     """Define a specific TestCase reltive to OrangeWorkflow"""
#
#     @classmethod
#     def setUpClass(cls):
#         TestCaseQt.setUpClass()
#         super(OrangeWorflowTest, cls).setUpClass()
#         super(OrangeWorflowTest, cls).init(cls)
#
#     @classmethod
#     def tearDownClass(cls):
#         cls.app.processEvents()
#         # make sure no events are resting
#         super(OrangeWorflowTest, cls).tearDown(cls)
#         TestCaseQt.tearDownClass()
#
#     def tearDown(self):
#         import time
#
#         while self.app.hasPendingEvents():
#             time.sleep(0.1)
#             self.app.processEvents()
#             self.processOrangeEventsStack()
#             time.sleep(0.1)
#             self.app.processEvents()
#
#         TestCase.tearDown(self)



# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""Some widget construction to check if a sample moved"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/03/2018"

from silx.gui import qt
from collections import OrderedDict
from silx.gui.widgets.TableWidget import TableWidget
from silx.io.url import DataUrl
import enum
import functools
import logging
import os

logger = logging.getLogger(__file__)


class ColumnMode(enum.Enum):
    SINGLE = ("single",)
    SINGLE_OR_NONE = ("single_or_none",)
    MULTIPLE = ("multiple",)
    MULTIPLE_OR_NONE = ("multiple_or_none",)


_DEFAULT_COLUMNS = OrderedDict(
    {("img B", ColumnMode.SINGLE), ("img A", ColumnMode.SINGLE)}
)


class UrlSelectionTable(TableWidget):
    """Table used to select the color channel to be displayed for each"""

    sigSelectionChanged = qt.Signal(dict)
    """Signal emitted when the a image selection changed
    Contains column id and the list of url selected (or None if none selected)
    """

    def __init__(self, parent=None, columns=_DEFAULT_COLUMNS):
        TableWidget.__init__(self, parent)
        self.selection_columns = columns
        self.clear()

    def clear(self):
        qt.QTableWidget.clear(self)
        self.setRowCount(0)
        self._column_index = OrderedDict({"url": 0})
        for i, column in enumerate(self.selection_columns):
            self._column_index[column] = i + 1

        self.setColumnCount(len(self._column_index))
        self.setHorizontalHeaderLabels(list(self._column_index.keys()))
        self.verticalHeader().hide()
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)

        self.setSortingEnabled(True)
        self._checkBoxes = {}
        self._url_path_to_url = {}

    def setUrls(self, urls):
        self.clear()
        self.setRowCount(len(urls))
        for iUrl, url in enumerate(urls):
            self.addUrl(url=url, row=iUrl, resize=False)
        self.sortItems(0)
        self.resizeColumnsToContents()

    def addUrl(self, url, row=None, resize=False, **kwargs):
        """

        :param url:
        :param args:
        :return: index of the created items row
        :rtype int
        """
        assert isinstance(url, DataUrl)
        if row is None:
            row = self.rowCount()
            self.setRowCount(row + 1)

        _item = qt.QTableWidgetItem()
        _item.setText(os.path.basename(url.path()))
        _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
        self.setItem(row, self._column_index["url"], _item)

        self._checkBoxes[url.path()] = {}
        # work for now since we are in a 'single' edf mode but will fail in the
        # future
        self._url_path_to_url[url.path()] = url
        for column_name in self.selection_columns:
            widgetImg = qt.QRadioButton(parent=self)
            widgetImg.setAutoExclusive(False)
            self.setCellWidget(row, self._column_index[column_name], widgetImg)
            callbackImg = functools.partial(self._selection_changed, column_name, url)
            widgetImg.toggled.connect(callbackImg)
            self._checkBoxes[url.path()][column_name] = widgetImg
        if resize is True:
            self.resizeColumnsToContents()
        return row

    def _selection_changed(self, column_name, url, toggle):
        column_mode = self.selection_columns[column_name]
        if toggle is True and column_mode in (
            ColumnMode.SINGLE,
            ColumnMode.SINGLE_OR_NONE,
        ):
            for tmp_url in self._checkBoxes:
                for tmp_column in self._checkBoxes[tmp_url]:
                    self._checkBoxes[tmp_url][tmp_column].blockSignals(True)
            self._set_column_selection(column=column_name, selection=[url])
            for tmp_url in self._checkBoxes:
                for tmp_column in self._checkBoxes[tmp_url]:
                    self._checkBoxes[tmp_url][tmp_column].blockSignals(False)

        # force a selection if need to have at least one and selection is empty
        # for now
        if toggle is False:
            if (
                column_mode
                in (
                    ColumnMode.SINGLE,
                    ColumnMode.MULTIPLE,
                )
                and self._has_no_element(column_name)
            ):
                for tmp_url in self._checkBoxes:
                    for tmp_column in self._checkBoxes[tmp_url]:
                        self._checkBoxes[tmp_url][tmp_column].blockSignals(True)
                selection = [url]
                self._set_column_selection(column=column_name, selection=selection)
                for tmp_url in self._checkBoxes:
                    for tmp_column in self._checkBoxes[tmp_url]:
                        self._checkBoxes[tmp_url][tmp_column].blockSignals(False)

    def _has_no_element(self, column):
        for url_path in self._checkBoxes:
            if self._checkBoxes[url_path][column].isChecked():
                return False
        return True

    def _updatecheckBoxes(self, whichImg, name):
        old = self.blockSignals(True)

        assert name in self._checkBoxes
        assert whichImg in self._checkBoxes[name]
        if self._checkBoxes[name][whichImg].isChecked():
            for radioUrl in self._checkBoxes:
                if radioUrl != name:
                    self._checkBoxes[radioUrl.path()][whichImg].setChecked(False)
        self.blockSignals(old)

    def getSelectedUrls(self, name=None):
        """

        :return: url selected for the requested name. Or full selection if no
                 name selected.
        """
        selection = {}
        for _name in self.selection_columns:
            selection[_name] = None

        for radioUrl in self._checkBoxes:
            for _name in self.selection_columns:
                if self._checkBoxes[radioUrl][_name].isChecked():
                    if selection[_name] is None:
                        selection[_name] = []
                    selection[_name].append(self._url_path_to_url[radioUrl])
        if name is None:
            return selection
        else:
            return selection[name]

    def setSelectedUrls(self, selection):
        """

        :param ddict: key: image url, values: list of active channels
        """
        assert isinstance(selection, dict)
        for name, sel in selection.items():
            assert isinstance(sel, (type(None), list, tuple))
            self._set_column_selection(column=name, selection=sel)
        self._signal_selection_changed()

    def _clear_column(self, column):
        """

        :param str column:
        """
        old = self.blockSignals(True)
        for radioUrl in self._checkBoxes:
            self._checkBoxes[radioUrl][column].setChecked(False)
        self.blockSignals(old)

    def _set_column_selection(self, column, selection):
        """

        :param str column: id of the column
        :param Union[None, list]: None or list of DataUrl
        """
        self._clear_column(column=column)
        old = self.blockSignals(True)
        if selection is not None:
            assert isinstance(selection, (list, tuple))
            for sel in selection:
                if sel is not None:
                    assert isinstance(sel, DataUrl)
                    self._checkBoxes[sel.path()][column].setChecked(True)
        self.blockSignals(old)

    def _signal_selection_changed(self):
        self.sigSelectionChanged.emit(self.getSelectedUrls())

    def removeUrl(self, url):
        raise NotImplementedError("")


class UrlSelectionDialog(qt.QDialog):
    """Embed the UrlSelectionWidget into a QDialog"""

    _sizeHint = qt.QSize(500, 500)

    def __init__(self, columns, parent=None):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.widget = UrlSelectionTable(columns=columns, parent=self)
        self.layout().addWidget(self.widget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        # connect signal / SLOT
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)

        # expose API
        self.setUrls = self.widget.setUrls
        self.addUrl = self.widget.addUrl
        self.setSelection = self.widget.setSelectedUrls
        self.getSelection = self.widget.getSelectedUrls

    def sizeHint(self):
        """Return a reasonable default size for usage in :class:`PlotWindow`"""
        return self._sizeHint



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/02/2017"

"""Module defining the dition of a single H5Struct"""

import logging
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructsEditor

logger = logging.getLogger(__name__)


class H5StructEditor(H5StructsEditor):
    """simple class which link an interface with an h5 struct used for pyhst"""

    def __init__(self, structID):
        self.structID = structID
        H5StructsEditor.__init__(self, (self.structID,))

    def linkComboboxWithH5Variable(
        self,
        qcombobox,
        h5ParamName,
        fitwithindex=False,
        dic=None,
        filters=None,
        setDefault=True,
    ):
        """

        :param qcombobox: the QCombobox in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param fitwithindex: true if we want to store the index in the h5File,
            otherwise store the string
        :param filters: filter to apply on the string to get the requested
            value
        """
        return H5StructsEditor.linkComboboxWithH5Variable(
            self,
            qcombobox=qcombobox,
            structID=self.structID,
            h5ParamName=h5ParamName,
            fitwithindex=fitwithindex,
            dic=dic,
            filters=filters,
            setDefault=setDefault,
        )

    def linkCheckboxWithH5Variable(self, qcheckbox, h5ParamName, invert=False):
        """

        :param qcheckbox: the QCheckBox in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param invert: true if we will not store the value state == checked but
            state == unchecked
        :param structID: the structID associated with the param ID
        """
        return H5StructsEditor.linkCheckboxWithH5Variable(
            self,
            qcheckbox=qcheckbox,
            structID=self.structID,
            h5ParamName=h5ParamName,
            invert=invert,
        )

    def LinkLineEditWithH5Variable(self, qlineedit, h5ParamName, h5paramtype=str):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        return H5StructsEditor.LinkLineEditWithH5Variable(
            self,
            qlineedit=qlineedit,
            h5ParamName=h5ParamName,
            structID=self.structID,
            h5paramtype=h5paramtype,
        )

    def LinkSelectionLineEditWithH5Variable(
        self, qlineedit, h5ParamName, h5paramtype=str
    ):
        """

        :param qlineedit: the QLineEdit in the interface to edit/display the
            given h5ParamName
        :param h5ParamName: the parameter ID in the H5 structure.
        :param h5ParamName: the parameter ID in the H5 structure.
        """
        return H5StructsEditor.LinkSelectionLineEditWithH5Variable(
            self,
            qlineedit=qlineedit,
            h5ParamName=h5ParamName,
            structID=self.structID,
            h5paramtype=h5paramtype,
        )

    def linkGroupWithH5Variable(self, group, h5ParamName, getter, setter):
        return H5StructsEditor.linkGroupWithH5Variable(
            self,
            group=group,
            structID=self.structID,
            h5ParamName=h5ParamName,
            getter=getter,
            setter=setter,
        )



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"

from tomwer.gui.stackplot import ImageFromFile
from tomwer.gui.stacks import SliceStack, RadioStack
from tomwer.test.utils import UtilsTest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
import logging
import time
import pytest
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.test.utils import skip_gui_test
import weakref

_qapp = QApplicationManager()

logging.disable(logging.INFO)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSliceStack(TestCaseQt):
    """unit test for the :class:_ImageStack widget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = SliceStack()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    def tearDown(self):
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        TestCaseQt.tearDown(self)

    def test(self):
        """Make sur the addLeaf and clear functions are working"""
        self._widget.setLoadingMode("load when show requested")
        folder = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self.assertTrue(self._widget._viewer.getActiveImage() is None)

        self.assertEqual(len(self._widget._scans), 0)
        self._widget.addLeafScan(folder)
        self.assertEqual(len(self._widget._scans), 1)
        # some delay to wait for the ImageLoaderThread
        time.sleep(0.3)
        _qapp.processEvents()
        self.assertTrue(self._widget._viewer.getActiveImage() is not None)
        self._widget.clear()
        self.assertEqual(len(self._widget._scans), 0)
        self.assertTrue(self._widget._viewer.getActiveImage() is None)

        self._widget.addLeafScan(folder)
        self.assertEqual(len(self._widget._scans), 1)
        # some delay to wait for the ImageLoaderThread
        time.sleep(0.3)
        _qapp.processEvents()
        self.assertTrue(self._widget._viewer.getActiveImage() is not None)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestRadioStack(TestCaseQt):
    """Test for the RadioStack"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = RadioStack()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.spinBox = weakref.ref(self._widget._viewer._qspinbox)
        self.slider = weakref.ref(self._widget._viewer._qslider)

    def tearDown(self):
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        TestCaseQt.tearDown(self)

    def _waitImages(self):
        _qapp.processEvents()
        time.sleep(0.2)
        _qapp.processEvents()

    def testASAPMode(self):
        self._widget.setLoadingMode("load ASAP")
        folder = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self._widget.setForceSync(True)
        self._widget.addLeafScan(folder)
        # wait for all images to be loaded
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (3600, 0, 0)
        )

        self.spinBox().setValue(20)
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (3600, 0, 0)
        )
        self.slider().setValue(10)
        _qapp.processEvents()
        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (3600, 0, 0)
        )

    def testLazyLoadingMode(self):
        self._widget.setLoadingMode("lazy loading")
        folder = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self._widget.addLeafScan(folder)
        self._waitImages()
        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (1, 0, 3599)
        )

        self.spinBox().setValue(20)
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (1, 0, 3599)
        )
        self.slider().setValue(10)
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (1, 0, 3599)
        )

        self.spinBox().editingFinished.emit()
        self._waitImages()
        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (2, 0, 3598)
        )

    def testOnShowLoading(self):
        self._widget.setLoadingMode("load when show requested")
        folder = UtilsTest.getEDFDataset("D2_H2_T2_h_")
        self._widget.addLeafScan(folder)
        # wait for all images to be loaded
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (1, 0, 3599)
        )

        self.spinBox().setValue(20)
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (2, 0, 3598)
        )
        self.slider().setValue(10)
        self._waitImages()

        self.assertEqual(
            self._getStatsLoadedImage(self._widget._viewer.images), (3, 0, 3597)
        )

    def _getStatsLoadedImage(self, images):
        """

        :param images:
        :return: nLoaded, NLoading, NNotLoaded
        """
        nLoaded = nLoading = nNotLoaded = 0
        for imgIndex, img in images._images.items():
            assert isinstance(img, ImageFromFile)
            nLoaded += int(img._status == "loaded")
            nLoading += int(img._status == "loading")
            nNotLoaded += int(img._status == "not loaded")
        return (nLoaded, nLoading, nNotLoaded)



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain utils for score process
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "27/10/2021"


from distributed import client
from tomwer.core.cluster.cluster import (
    SlurmClusterConfiguration,
    patch_worker_info_to_worker,
)
from tomwer.core.process.reconstruction.nabu.target import Target
from tomwer.core.process.reconstruction.nabu.utils import _NabuPhaseMethod
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.slurm import is_slurm_available
from silx.utils.enum import Enum as _Enum
from .slurm import _exec_nabu_on_slurm
from typing import Iterable, Optional
from typing import Union
import logging
import subprocess
from . import settings
from . import utils
from time import sleep
import os

_logger = logging.getLogger(__name__)
try:
    from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
        ChunkedReconstructor,
    )
except ImportError as e:
    _logger.error(e)
    has_nabu = False
else:
    has_nabu = True


class NabuOutputFileFormat(_Enum):
    # NPY = 'npy'
    # NPZ = 'npz'
    TIFF = "tiff"
    HDF5 = "hdf5"
    JP2K = "jp2k"
    EDF = "edf"

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            value = value.lstrip(".")
        return super().from_value(value)


def get_file_format(file_str):
    extension = os.path.splitext(file_str.lower())[-1]
    extension = extension.lstrip(".")
    if extension in ("tiff", "tif"):
        return NabuOutputFileFormat.TIFF
    elif extension in ("hdf5", "hdf", "h5"):
        return NabuOutputFileFormat.HDF5
    elif extension in ("jp2k", "jpg2k"):
        return NabuOutputFileFormat.JP2K
    elif extension in ("edf",):
        return NabuOutputFileFormat.EDF
    else:
        raise ValueError(f"Unrecognized file extension {extension} from {file_str}")


class ResultsRun:
    """
    Base class of results for nabu
    """

    def __init__(self, success, config) -> None:
        self.__success = success
        self.__config = config

    @property
    def success(self) -> bool:
        return self.__success

    @property
    def config(self) -> dict:
        return self.__config

    def __str__(self) -> str:
        return f"result from nabu run: {'succeed' if self.success else 'failed'} with \n - config:{self.config} \n"


class ResultsWithStd(ResultsRun):
    """Nabu result with std"""

    def __init__(self, std_out, std_err, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__std_err = std_err
        self.__std_out = std_out

    @property
    def std_out(self) -> str:
        return self.__std_out

    @property
    def std_err(self) -> str:
        return self.__std_err

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n {self.std_out} \n {self.std_err}"
        return res


class ResultsLocalRun(ResultsWithStd):
    """Nabu result when run locally.
    If this is the case we should be able to retrieve directly the results urls"""

    def __init__(
        self,
        results_urls: tuple,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if not isinstance(results_urls, Iterable):
            raise TypeError(
                f"results_urls is expected to be an Iterable not {type(results_urls)}"
            )
        self.__results_urls = results_urls

    @property
    def results_urls(self) -> tuple:
        return self.__results_urls

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n - result urls: {self.results_urls}"
        return res


class ResultSlurmRun(ResultsWithStd):
    """Nabu result when run on slurm. on this case we expect to get a future and a distributed client"""

    def __init__(
        self,
        future_slurm_jobs: tuple,
        client,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.__future_slurm_jobs = future_slurm_jobs
        self.client = client

    @property
    def future_slurm_jobs(self):
        return self.__future_slurm_jobs

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n - future job slurms: {self.future_slurm_jobs} \n - client: {client}"
        return res


class _NabuBaseReconstructor:
    """
    Base class to submit a job to nabu
    """

    TIMEOUT_SLURM_JOB_SUBMISSION = 30
    """Timeout when submit a job to slurm cluster. In second"""

    def __init__(
        self,
        scan: TomwerScanBase,
        dry_run: bool,
        target: Target,
        cluster_config: Optional[Union[dict, SlurmClusterConfiguration]],
        process_name: str,
    ) -> None:
        self._scan = scan
        self._target = Target.from_value(target)
        self._dry_run = dry_run
        self._process_name = process_name
        if isinstance(cluster_config, SlurmClusterConfiguration):
            self._cluster_config = cluster_config
        elif isinstance(cluster_config, dict):
            self._cluster_config = SlurmClusterConfiguration.from_dict(cluster_config)
        elif cluster_config is None:
            self._cluster_config = None
        else:
            raise TypeError(
                f"cluster config is expected to be a dict or an instance of {SlurmClusterConfiguration}. Not {type(cluster_config)}"
            )

    @property
    def scan(self):
        return self._scan

    @property
    def target(self):
        return self._target

    @property
    def cluster_config(self):
        return self._cluster_config

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def process_name(self):
        return self._process_name

    def only_create_config_file(self):
        """Should we run the reconstruction or only create the configuration file"""
        return False

    def run(self) -> Iterable:
        """
        run the requested slices.

        :return: Iterable of ResultsRun.
        """
        raise NotImplementedError("Base class")

    def _process_config(
        self,
        config_to_dump: dict,
        config_file: str,
        file_format: str,
        start_z: Optional[int],
        end_z: Optional[int],
        info: Optional[str],
        process_name: str,
    ):
        """
        process provided configuration

        :param str info:
        """
        if self.dry_run is True or self.only_create_config_file():
            return ResultsRun(
                success=True,
                config=config_to_dump,
            )
        elif self.target is Target.LOCAL:
            _logger.info(f"run {info} for {self.scan} with {config_to_dump}")
            return self._run_nabu_locally(
                conf_file=config_file,
                file_format=file_format,
                config_to_dump=config_to_dump,
                start_z=start_z,
                end_z=end_z,
            )
        elif self.target is Target.SLURM:
            _logger.info(
                f"run {info} on slurm for {self.scan.path} with {config_to_dump}"
            )
            return self._run_nabu_on_slurm(
                conf_file=config_file,
                config_to_dump=config_to_dump,
                cluster_config=self.cluster_config.to_dict(),
                start_z=start_z,
                end_z=end_z,
                process_name=process_name,
                info=info,
            )
        else:
            raise ValueError(f"{self.target} is not recognized as a valid target")

    def _run_nabu_locally(
        self,
        conf_file: str,
        file_format: str,
        config_to_dump: dict,
        start_z: int,
        end_z: int,
    ) -> ResultsLocalRun:
        """
        run locally nabu for a single configuration file.

        :param str conf_file: path to the nabu .cfg file
        :param str file_format: format of the generated file
        :param dict config_to_dump: configuration saved in the .cfg as a dictionary
        :return: results of the local run
        :rtype: ResultsLocalRun
        """
        if not has_nabu:
            raise ImportError("Fail to import nabu")
        command = " ".join(
            ("python", "-m", settings.NABU_FULL_FIELD_APP_PATH, conf_file)
        )
        _logger.info('call nabu from "{}"'.format(command))
        process = subprocess.Popen(
            command,
            shell=True,
            cwd=self.scan.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        outs, errs = process.communicate()

        recons_urls = utils.get_recons_urls(
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            slice_index=None,
            scan=self.scan,
            file_format=file_format,
            start_z=start_z,
            end_z=end_z,
        )
        return ResultsLocalRun(
            success=len(recons_urls) > 0,
            results_urls=recons_urls,
            std_out=outs,
            std_err=errs,
            config=config_to_dump,  # config_slices,
        )

    def _run_nabu_on_slurm(
        self,
        conf_file: str,
        config_to_dump: dict,
        cluster_config: dict,
        start_z: int,
        end_z: int,
        process_name: str,
        info: str,
    ) -> ResultSlurmRun:
        """
        Run a nabu reconstruction on slurm of a single configuration

        :return: results of the slurm run
        :rtype: ResultSlurmRun
        """
        if not isinstance(conf_file, str):
            raise TypeError(f"conf_file is expected to be a strg not {type(conf_file)}")
        if not isinstance(config_to_dump, dict):
            raise TypeError(
                f"config_to_dump is expected to be a strg not {type(config_to_dump)}"
            )
        if not is_slurm_available():
            raise RuntimeError("slurm not available")
        if not isinstance(cluster_config, dict):
            raise ValueError(
                f"cluster config is expected to be a dict not {type(cluster_config)}"
            )
        from distributed import Client
        from tomwer.core.cluster import SlurmClusterManager

        # create slurm cluster
        project_name = cluster_config.get(
            "project_name", "tomwer_{scan}_-_{process}_-_{info}"
        )
        project_name = project_name.format(
            scan=str(self.scan), process=process_name, info=info
        )
        # project name should not contain any spaces as it will be integrated in a script and interpreted.
        project_name = project_name.replace(" ", "_")

        cluster = SlurmClusterManager().get_cluster(
            cluster_config,
            project_name=project_name,
        )
        client = Client(cluster)

        # do not rely on "cluster.workers", as it reports status=='running' when it's not actually ready
        timeout = self.TIMEOUT_SLURM_JOB_SUBMISSION
        loop_duration = 0.5
        while len(client.has_what()) != int(cluster_config["n_workers"]):
            timeout -= loop_duration
            if timeout == 0:
                raise TimeoutError(
                    f"unable to submit job to {cluster}. No workers spawn after {self.TIMEOUT_SLURM_JOB_SUBMISSION} seconds"
                )
            sleep(loop_duration)

        patch_worker_info_to_worker(client)
        # submit job
        future_slurm_job = client.submit(
            _exec_nabu_on_slurm,
            conf_file,
            self.scan.path,
        )

        callbacks = self._get_futures_slurm_callback(config_to_dump)
        assert isinstance(
            callbacks, tuple
        ), f"callbacks is expected to an instance of tuple and not {type(callbacks)}"
        for callback in callbacks:
            future_slurm_job.add_done_callback(callback.process)

        return ResultSlurmRun(
            success=True,
            config=config_to_dump,
            future_slurm_jobs=(future_slurm_job,),
            std_out=None,
            std_err=None,
            client=client,
        )

    def _get_futures_slurm_callback(self, config_to_dump) -> tuple:
        """Return a tuple a potential callback to be launch once the future is done"""
        return tuple()

    def _treateOutputSliceConfig(self, config):
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist
        """
        pag = False
        db = None
        if "phase" in config:
            if (
                "method" in config["phase"]
                and config["phase"]["method"] == _NabuPhaseMethod.PAGANIN.value
            ):
                pag = True
                if "delta_beta" in config["phase"]:
                    db = round(float(config["phase"]["delta_beta"]))
        if "output" in config:
            _file_name = self._get_file_basename_reconstruction(pag=pag, db=db)
            config["output"]["file_prefix"] = _file_name
            location = config["output"].get("location", None)
            if location not in ("", None):
                location = self.format_output_location(location, scan=self.scan)
                # if user specify the location
                if not os.path.isdir(config["output"]["location"]):
                    os.makedirs(location)
            else:
                # otherwise default location will be the data root level
                location = self.scan.path
            config["output"]["location"] = location
        return config

    def _get_file_basename_reconstruction(self, pag, db):
        """return created file base name"""
        raise NotImplementedError("Base class")

    @staticmethod
    def format_output_location(location, scan: TomwerScanBase):
        """
        format possible keys from the location like {scan_dir} or {scan_path}

        :param location:
        :param scan:
        :return:
        """
        if scan is None:
            _logger.warning("scan is !none, enable to format the nabu output location")

        keywords = {
            "scan_dir_name": scan.scan_dir_name(),
            "scan_basename": scan.scan_basename(),
            "scan_parent_dir_basename": scan.scan_parent_dir_basename(),
        }
        for keyword, value in keywords.items():
            if value is None:
                continue
            try:
                location = location.format(**{keyword: value})
            except KeyError:
                # then this mean scan_dir has not been provided
                pass
        return location



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
Contains processing relative to a `Slurmcluster`
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "12/10/2021"


from distributed.client import Client
from tomwer.core.utils.Singleton import singleton as _singleton
from tomwer.core.settings import SlurmSettings as _SlurmSettings
from dask_jobqueue import SLURMCluster as _SLURMCluster
import os


class SlurmClusterConfiguration:
    """Object shipping the configuration of a slurm cluster"""

    def __init__(
        self,
        n_cores=_SlurmSettings.N_CORES_PER_WORKER,
        n_workers=_SlurmSettings.N_WORKERS,
        memory=_SlurmSettings.MEMORY_PER_WORKER,
        queue=_SlurmSettings.QUEUE,
        n_gpus=_SlurmSettings.N_GPUS_PER_WORKER,
        project_name=_SlurmSettings.PROJECT_NAME,
        walltime=_SlurmSettings.DEFAULT_WALLTIME,
        python_venv=_SlurmSettings.PYTHON_VENV,
        port_range=_SlurmSettings.PORT_RANGE,
        dashboard_port=_SlurmSettings.DASHBOARD_PORT,
    ) -> None:
        self._n_cores = n_cores
        self._n_workers = n_workers
        self._memory = memory
        self._queue = queue
        self._n_gpus = n_gpus
        self._project_name = project_name
        self._walltime = walltime
        self._python_venv = python_venv
        self._port_range = port_range
        self._dashboard_port = dashboard_port

    @property
    def n_cores(self):
        return self._n_cores

    @property
    def n_workers(self):
        return self._n_workers

    @property
    def memory(self):
        return self._memory

    @property
    def queue(self):
        return self._queue

    @property
    def n_gpus(self):
        return self._n_gpus

    @property
    def project_name(self):
        return self._project_name

    @property
    def walltime(self):
        return self._walltime

    @property
    def python_venv(self):
        return self._python_venv

    @property
    def port_range(self) -> tuple:
        """port range as (start:int, strop:int, step: int)"""
        return self._port_range

    @property
    def dashboard_port(self):
        return self._dashboard_port

    def to_dict(self) -> dict:
        return {
            "cores": self.n_cores,
            "n_workers": self.n_workers,
            "memory": self.memory,
            "queue": self.queue,
            "n_gpus": self.n_gpus,
            "project_name": self.project_name,
            "walltime": self.walltime,
            "python_venv": self.python_venv,
            "port_range": self.port_range,
            "dashboard_port": self.dashboard_port,
        }

    @staticmethod
    def from_dict(dict_: dict):
        return SlurmClusterConfiguration().load_from_dict(dict_=dict_)

    def load_from_dict(self, dict_: dict):
        if "cores" in dict_:
            self._n_cores = dict_["cores"]
        if "n_workers" in dict_:
            self._n_workers = dict_["n_workers"]
        if "memory" in dict_:
            self._memory = dict_["memory"]
        if "queue" in dict_:
            self._queue = dict_["queue"]
        if "n_gpus" in dict_:
            self._n_gpus = dict_["n_gpus"]
        if "project_name" in dict_:
            self._project_name = dict_["project_name"]
        if "walltime" in dict_:
            self._walltime = dict_["walltime"]
        if "python_venv" in dict_:
            self._python_venv = dict_["python_venv"]
        if "port_range" in dict_:
            self._port_range = dict_["port_range"]
        if "dashboard_port" in dict_:
            self._dashboard_port = dict_["dashboard_port"]
        return self


@_singleton
class SlurmClusterManager:
    def __init__(self) -> None:
        self._clusters = {}
        # associate at each cluster configuration a SLURMCluster

    def _config_dict_to_tuple(self, config: dict):
        return (
            config["cores"],
            config["n_workers"],
            config["memory"],
            config["n_gpus"],
            config["queue"],
            config["port_range"],
            config["dashboard_port"],
        )

    def get_cluster(self, config: dict, project_name: str) -> _SLURMCluster:
        n_gpus = config.get("n_gpus", None)
        job_extra = []
        if n_gpus not in (0, None):
            job_extra = ["--gres=gpu:{}".format(n_gpus)]

        import socket
        from contextlib import closing

        # from stackoverflow: https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number
        def find_free_port():
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
                s.bind(("", 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return s.getsockname()[1]

        def dict_to_SLURMClusterParams(dict_):
            """convert a dict to valid parameter input"""
            python_venv = dict_.get("python_venv", None)
            host_port = find_free_port()
            dashboard_port = host_port
            while dashboard_port == host_port:
                dashboard_port = find_free_port()

            res = {
                "memory": f"{dict_['memory']}GB",
                "cores": dict_["cores"],
                "project": project_name,
                "queue": dict_["queue"],
                "job_extra": job_extra,
                "processes": 1,
                "scheduler_options": {
                    "host": "0.0.0.0:" + str(host_port),
                    "dashboard_address": "0.0.0.0:" + str(dashboard_port),
                },
                "walltime": dict_["walltime"],
            }
            if python_venv not in (None, ""):
                # warning: we need to provide both the executable and the
                # activation script because we will call some subprocess with python -m
                # anyway this is a better practice to have a 'logical' context.
                python_exe = os.path.join(python_venv, "bin", "python")
                python_activate = os.path.join(python_venv, "bin", "activate")
                res.update(
                    {
                        "python": python_exe,
                        "env_extra": ["source {}".format(python_activate)],
                    }
                )
            return res

        cluster_constructor_params = dict_to_SLURMClusterParams(config)
        new_cluster = _SLURMCluster(**cluster_constructor_params)
        new_cluster.scale(config["n_workers"])

        return new_cluster


def patch_worker_info_to_worker(client: Client):
    """simple function to patch the client with some extra information
    like the job_id
    """

    def get_job_id():
        return int(os.environ.get("SLURM_JOB_ID", -1))

    job_ids = {}
    for worker in client.has_what().keys():
        fut = client.submit(get_job_id, workers=[worker])
        job_ids[worker] = fut.result()

    client.job_ids = job_ids



class DeviceProxy:
    def ScanNumber(self):
        pass

    def State(self):
        pass

    def SequenceScanNumber(self):
        pass

    def SavingFile(self):
        pass

    def ScanTitle(self):
        pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "20/04/2020"


import os
import threading

from contextlib import AbstractContextManager
from tomwer.core.utils.Singleton import singleton


@singleton
class FileLockerManager:
    """Insure that for each file we will provide at most one locker"""

    def __init__(self):
        self.__lockers = {}

    def get_locker(self, file_):
        file_ = os.path.abspath(file_)
        if file_ not in self.__lockers:
            self.__lockers[file_] = threading.RLock()
        return self.__lockers[file_]

    def clear_locker(self, file_):
        if file_ in self.__lockers:
            del self.__lockers[file_]


class FileLockerContext(AbstractContextManager):
    """provide a context manager for a FileLocker"""

    def __init__(self, lock):
        self.__lock = lock

    def __enter__(self):
        self.__lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__lock.release()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/05/2018"

from orangewidget import gui
from orangewidget.widget import Output, Input, OWBaseWidget
from tomwer.web.client import OWClient
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.scanselectorwidget import ScanSelectorWidget
from tomwer.core.scan.scanbase import TomwerScanBase
from orangewidget.settings import Setting
import tomwer.core.process.control.scanvalidator
import logging

logger = logging.getLogger(__name__)


class DataSelectorOW(OWBaseWidget, OWClient, openclass=True):
    name = "data selector"
    id = "orange.widgets.tomwer.scanselector"
    description = (
        "List all received scan. Then user can select a specific"
        "scan to be passed to the next widget."
    )
    icon = "icons/scanselector.svg"
    priority = 42
    keywords = ["tomography", "selection", "tomwer", "folder"]

    ewokstaskclass = tomwer.core.process.control.scanvalidator._ScanValidatorPlaceHolder

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _scanIDs = Setting(list())

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """ """
        OWBaseWidget.__init__(self, parent)
        OWClient.__init__(self)

        self.widget = ScanSelectorWidget(parent=self)
        self._loadSettings()

        self.widget.sigUpdated.connect(self._updateSettings)
        self.widget.sigSelectionChanged.connect(self.changeSelection)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        # expose API
        self.setActiveScan = self.widget.setActiveScan
        self.selectAll = self.widget.selectAll
        self.n_scan = self.widget.n_scan
        self.add = self.widget.add

    @Inputs.data
    def addScan(self, scan):
        if scan is not None:
            self.widget.add(scan)

    def changeSelection(self, list_scan):
        if list_scan:
            for scan_id in list_scan:
                if scan_id in self.widget.dataList._scanIDs:
                    scan = self.widget.dataList._scanIDs[scan_id]
                    assert isinstance(scan, TomwerScanBase)
                    self.Outputs.data.send(scan)
                else:
                    logger.error("%s not found in scan ids" % scan_id)

    def send(self):
        """send output signals for each selected items"""
        sItem = self.widget.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [_item.text() for _item in sItem]
            self.changeSelection(list_scan=selection)

    def _loadSettings(self):
        for scan in self._scanIDs:
            assert isinstance(scan, str)
            if "@" in scan:
                entry, file_path = scan.split("@")
                hdf5_tomo_scan = HDF5TomoScan(entry=entry, scan=file_path)
                self.addScan(hdf5_tomo_scan)
            else:
                self.addScan(scan)

    def _updateSettings(self):
        self._scanIDs = []
        for scan in self.widget.dataList.items:
            self._scanIDs.append(scan)



import subprocess
import logging

_logger = logging.getLogger(__name__)


def check_n_gpu():
    try:
        from pycuda import driver

        driver.init()
        n_gpus = driver.Device.count()
    except Exception as e:
        raise e
    else:
        if n_gpus == 0:
            raise RuntimeError("No gpu detected to run nabu reconstruction")
        else:
            _logger.info(f"{n_gpus} detected by pycuda")


def _exec_nabu_on_slurm(conf_file: str, scan_path: str) -> tuple:
    check_n_gpu()

    NABU_FULL_FIELD_APP_PATH = "nabu.app.reconstruct"
    # need to be executed in his own context
    command = " ".join(("python", "-m", NABU_FULL_FIELD_APP_PATH, conf_file))
    process = subprocess.Popen(
        command,
        shell=True,
        cwd=scan_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    res = process.communicate()
    return res



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/03/2020"

from silx.gui import qt
from orangewidget import widget, gui
from tomwer.gui.control.datalist import BlissHDF5DataListMainWindow
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.web.client import OWClient
from tomwer.core.process.control.nxtomomill import NxTomomillProcess
import tomwer.core.process.control.nxtomomill
from orangewidget.settings import Setting
from orangewidget.widget import Output, Input
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.blissscan import BlissScan
from nxtomomill import converter as nxtomomill_converter

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config
import os
import logging

logger = logging.getLogger(__name__)


class InputDialogWithCache(qt.QDialog):
    """
    Input dialog for get missing information for nxtomomill translation
    Identical to a default QInputDialog adding a CheckBox to store answer.
    """

    def __init__(self, parent, scan, title="", desc="", cache_answer=True):
        qt.QDialog.__init__(self, parent)
        self.setWindowFlags(qt.Qt.MSWindowsFixedSizeDialogHint)

        self.setWindowTitle(title)
        self.setLayout(qt.QGridLayout())

        # entry
        self.layout().addWidget(qt.QLabel("entry:", self), 0, 0, 1, 1)
        self._entryQLE = qt.QLineEdit(scan.entry if scan else "", self)
        self._entryQLE.setReadOnly(True)
        self.layout().addWidget(self._entryQLE, 0, 1, 1, 1)

        # file name
        self.layout().addWidget(qt.QLabel("file:", self), 1, 0, 1, 1)
        self._fileQLE = qt.QLineEdit(scan.master_file if scan else "", self)
        self._fileQLE.setReadOnly(True)
        self.layout().addWidget(self._fileQLE, 1, 1, 1, 1)

        # label message
        self._label = qt.QLabel(desc, self)
        self.layout().addWidget(self._label, 2, 0, 1, 2)

        # input QLineEdit
        self._input = qt.QLineEdit("", self)
        self.layout().addWidget(self._input, 3, 0, 1, 2)

        # cache combobox
        self._cacheAnswerCB = qt.QCheckBox("keep answer in cache", self)
        self._cacheAnswerCB.setToolTip(
            "Will keep in memory the answer for the"
            " 'current cycle'. A cycle starts when "
            "you press 'send all' or "
            "'send selected' and ends when the last"
            " scan as been converted. The cache is "
            "reset after each cycle"
        )
        font = self._cacheAnswerCB.font()
        font.setPixelSize(10)
        self._cacheAnswerCB.setFont(font)
        self.layout().addWidget(self._cacheAnswerCB, 4, 0, 1, 1)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 5, 0, 1, 1)

        # buttons for validation
        self._buttons = qt.QDialogButtonBox(self)
        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons.setStandardButtons(types)

        self.layout().addWidget(self._buttons, 4, 0, 1, 2)

        # set up
        self._cacheAnswerCB.setChecked(cache_answer)

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)

    def setLabelText(self, txt):
        self._label.setText(txt)

    def getText(self):
        return self._input.text()

    def setScan(self, entry, file_path):
        self._entryQLE.setText(entry)
        self._fileQLE.setText(file_path)

    def cache_answer(self):
        return self._cacheAnswerCB.isChecked()

    def setBlissScan(self, entry, file_path):
        self._fileQLE.setText(file_path)
        self._entryQLE.setText(entry)

    def clear(self):
        self.setWindowTitle("")
        self._label.setText("")
        self._input.setText("")


class _NXTomomillInput:
    """
    callback provided to nxtomomill if an entry is missing.
    The goal is to ask the user the missing informations

    :param str entry: name of the entry missing
    :param str desc: description of the entry
    :return: user input or None
    """

    def __init__(self, parent=None, scan=None):
        self._dialog = InputDialogWithCache(parent=None, scan=scan)
        self._cache = {}
        """Cache to be used the user want to uase back answer for some question
        """

    def exec_(self, field, desc):
        self._dialog.clear()
        self._dialog.setWindowTitle(field)
        self._dialog.setLabelText(desc)

        if field in self._cache:
            return self._cache[field]
        else:
            if self._dialog.exec_():
                answer = self._dialog.getText()
                if self._dialog.cache_answer():
                    self._cache[field] = answer
                return answer
            else:
                return None

    def setBlissScan(self, entry, file_path):
        self._dialog.setScan(entry=entry, file_path=file_path)

    __call__ = exec_


class _OverwriteMessage(qt.QDialog):
    def __init__(self, parent, message=""):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self._label = qt.QLabel(message, self)
        self.layout().addWidget(self._label, 0, 0, 1, 2)

        self._canOverwriteAllCB = qt.QCheckBox("overwrite all", self)
        self._canOverwriteAllCB.setToolTip(
            "Will keep in memory the right to overwrite nxtomomill output "
            "files for the 'current cycle'. A cycle starts when you press "
            "'send all' or 'send selected' and ends when the last scan as been"
            " converted. The cache is reset after each cycle"
        )
        font = self._canOverwriteAllCB.font()
        font.setPixelSize(10)
        self._canOverwriteAllCB.setFont(font)
        self.layout().addWidget(self._canOverwriteAllCB, 2, 0, 1, 1)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 3, 0, 1, 1)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.No
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons, 8, 0, 1, 1)

        # signal / slot connection
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.No).clicked.connect(self.reject)

    def setText(self, msg):
        self._label.setText(msg)

    def canOverwriteAll(self):
        return self._canOverwriteAllCB.isChecked()


class NXTomomillOW(widget.OWBaseWidget, OWClient, openclass=True):
    """
    Widget to allow user to pick some bliss files and that will convert them
    to HDF5scan.
    """

    name = "nxtomomill"
    id = "orange.widgets.tomwer.control.NXTomomillOW.NXTomomillOW"
    description = (
        "Read a bliss .h5 file and extract from it all possible"
        "NxTomo. When validated create a TomwerBaseScan for each "
        "file and entry"
    )
    icon = "icons/nxtomomill.svg"
    priority = 120
    keywords = ["hdf5", "nexus", "tomwer", "file", "convert", "NXTomo", "tomography"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _scans = Setting(list())

    _nxtomo_cfg_file = Setting(str())
    # to keep bacard compatibility

    _static_input = Setting(dict())

    ewokstaskclass = tomwer.core.process.control.nxtomomill.NxTomomillProcess

    class Inputs:
        data = Input(name="bliss data", type=BlissScan, doc="bliss scan to be process")

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        OWClient.__init__(self)
        self.widget = BlissHDF5DataListMainWindow(parent=self)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        self.__request_input = True
        # do we ask the user for input if missing
        self._inputGUI = None
        """Gui with cache for missing field in files to be converted"""
        self._canOverwriteOutputs = False
        """Cache to know if we have to ask user permission for overwriting"""

        # expose API
        self.add = self.widget.add
        self.n_scan = self.widget.n_scan
        # alias used for the 'simple workflow' for now
        self.start = self._sendAll

        # connect signal / slot
        self.widget._sendButton.clicked.connect(self._sendAll)
        self.widget._sendSelectedButton.clicked.connect(self._sendSelected)
        self.widget.sigNXTomoCFGFileChanged.connect(self._saveNXTomoCfgFile)
        self.widget.sigUpdated.connect(self._updateSettings)
        # handle settings
        self._loadSettings()

    @property
    def request_input(self):
        return self.__request_input

    @request_input.setter
    def request_input(self, request):
        self.__request_input = request

    def _saveNXTomoCfgFile(self, cfg_file):
        """save the nxtomofile to the setttings"""
        self._nxtomo_cfg_file = cfg_file
        self._static_input["nxtomomill_cfg_file"] = cfg_file

    def _sendSelected(self):
        """Send a signal for selected scans found to the next widget"""
        self._inputGUI = _NXTomomillInput()
        # reset the GUI for input (reset all the cache for answers)
        self._canOverwriteOutputs = False
        for bliss_url in self.widget.datalist.selectedItems():
            entry, file_path = bliss_url.text().split("@")
            self._inputGUI.setBlissScan(entry=entry, file_path=file_path)
            self._process(bliss_url.text())

    def _sendAll(self):
        """Send a signal for each scan found to the next widget"""
        self._inputGUI = _NXTomomillInput()
        # reset the GUI for input (reset all the cache for answers)
        self._canOverwriteOutputs = False
        for bliss_url in self.widget.datalist.items:
            entry, file_path = bliss_url.split("@")
            self._inputGUI.setBlissScan(entry=entry, file_path=file_path)
            self._process(bliss_url)

    def _process(self, bliss_url):
        """

        :param str bliss_url: string at entry@file format
        """
        logger.processStarted("Start translate {} to NXTomo".format(str(bliss_url)))

        entry, file_path = bliss_url.split("@")
        bliss_scan = BlissScan(master_file=file_path, entry=entry, proposal_file=None)
        self._processBlissScan(bliss_scan)

    def _userAgreeForOverwrite(self, file_path):
        if self._canOverwriteOutputs:
            return True
        else:
            msg = _OverwriteMessage(self)
            text = "NXtomomill will overwrite \n %s. Do you agree ?" % file_path
            msg.setText(text)
            if msg.exec_():
                self._canOverwriteOutputs = msg.canOverwriteAll()
                return True
            else:
                return False

    @Inputs.data
    def treatBlissScan(self, bliss_scan):
        self._processBlissScan(bliss_scan)

    def _processBlissScan(self, bliss_scan):
        if bliss_scan is None:
            return
        output_file_path = NxTomomillProcess.deduce_output_file_path(
            bliss_scan.master_file,
            entry=bliss_scan.entry,
            outputdir=self.widget.getOutputFolder(),
        )
        # check user has rights to write on the folder
        dirname = os.path.dirname(output_file_path)
        if not os.access(dirname, os.W_OK):
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "You don't have write rights on {}. Unable to generate "
                "the nexus file associated to {}".format(dirname, str(bliss_scan))
            )
            msg.setWindowTitle("No rights to write")
            msg.setText(text)
            msg.show()
            return
        # check if need to overwrite the file
        elif os.path.exists(output_file_path):
            if not self._userAgreeForOverwrite(output_file_path):
                return

        configuration_file = self.widget.getCFGFilePath()
        if configuration_file in (None, ""):
            configuration = HDF5Config()
        else:
            try:
                configuration = HDF5Config.from_cfg_file(configuration_file)
            except Exception as e:
                logger.error(
                    "Fail to use configuration file {}."
                    "Error is {}. No conversion will be done."
                    "".format(configuration_file, e)
                )
                return
        # TODO: management of inputs from url: do we want to ignore urls
        # provided ? Or check if the entry receive fit
        # one provided (a root entry for now)
        # force some parameters
        configuration.input_file = bliss_scan.master_file
        configuration.output_file = output_file_path
        configuration.entries = (bliss_scan.entry,)
        configuration.single_file = False
        configuration.overwrite = True
        configuration.request_input = self.request_input
        configuration.file_extension = ".nx"
        try:
            convs = nxtomomill_converter.from_h5_to_nx(
                configuration=configuration,
                input_callback=self._inputGUI,
                progress=None,
            )
        except Exception as e:
            logger.processFailed(
                "Fail to convert from bliss file: %s to NXTomo."
                "Conversion error is: %s" % (bliss_scan, e)
            )
        else:
            # in the case of zserie we can have several scan for one entry
            for conv in convs:
                conv_file, conv_entry = conv
                scan_converted = HDF5TomoScan(scan=conv_file, entry=conv_entry)
                logger.processSucceed(
                    "{} has been translated to {}" "".format(bliss_scan, scan_converted)
                )
                self.Outputs.data.send(scan_converted)

    def _loadSettings(self):
        for scan in self._scans:
            assert isinstance(scan, str)
            if "@" in scan:
                try:
                    entry, file_path = scan.split("@")
                    self.widget.add(scan, configuration=self.widget.getHDF5Config())
                except Exception as e:
                    logger.error("Fail to add {}. Error is {}".format(scan, e))
            else:
                logger.warning("{} is an invalid link to a file".format(scan))
        if "nxtomomill_cfg_file" in self._static_input:
            nxtomo_cfg_file = self._static_input["nxtomomill_cfg_file"]
        else:
            nxtomo_cfg_file = self._nxtomo_cfg_file
        self.widget.setCFGFilePath(nxtomo_cfg_file)
        if "output_dir" in self._static_input:
            self.widget.setOutputFolder(self._static_input["output_dir"])

    def _updateSettings(self):
        self._scans = []
        for scan in self.widget.datalist.items:
            self._scans.append(scan)
        self._static_input["output_dir"] = self.widget.getOutputFolder()



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from silx.gui import qt
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructEditor
from tomwer.synctools.ftseries import QReconsParams, _QFTRP

logger = logging.getLogger(__name__)


class DisplayWidget(H5StructEditor, qt.QWidget):
    """
    Create the widget inside the of the Display tab

    :param _ReconsParam reconsparams: reconstruction parameters edited by the
                                      widget
    """

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructEditor.__init__(self, structID="FT")
        self._recons_params = None
        self.setReconsParams(reconsparams)
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.__buildShowProj())
        self.layout().addWidget(self.__buildShowSlice())
        self.layout().addWidget(self.__buildAngleOffset())

        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        self._makeConnection()

    def setReconsParams(self, recons_params):
        if isinstance(recons_params, QReconsParams):
            _recons_params = recons_params.ft
        elif isinstance(recons_params, _QFTRP):
            _recons_params = recons_params
        else:
            raise ValueError(
                "recons_params should be an instance of QReconsParam or _QFTRP"
            )

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = _recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def _update_params(self):
        """Update all parameter"""
        self.load(self._recons_params)

    def _makeConnection(self):
        self._qcbShowProj.toggled.connect(self._showProjChanged)
        self._qcbShowSlice.toggled.connect(self._showSliceChanged)
        self._qleAngleOffset.editingFinished.connect(self._angleOffsetValueChanged)
        self._qleAngleOffset.editingFinished.connect(self._angleOffsetChanged)

    def __buildShowProj(self):
        self._qcbShowProj = qt.QCheckBox(
            "show graphical proj during reconstruction", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbShowProj, "SHOWPROJ")
        # those option are not managed anymore, we should provide an
        # option 'on in imageJ' instead
        self._qcbShowProj.hide()
        return self._qcbShowProj

    def _showProjChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(parameter="SHOWPROJ", value=int(b))

    def __buildShowSlice(self):
        self._qcbShowSlice = qt.QCheckBox(
            "show graphical slice during reconstruction", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbShowSlice, "SHOWSLICE")
        # those option are not managed anymore, we should provide an
        # option 'on in imageJ' instead
        self._qcbShowSlice.hide()
        return self._qcbShowSlice

    def _showSliceChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(
                parameter="SHOWSLICE", value=int(b)
            )

    def __buildAngleOffset(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())
        widget.layout().addWidget(
            qt.QLabel("Final image rotation angle (degree):", parent=widget)
        )

        self._qleAngleOffset = qt.QLineEdit("", widget)
        validator = qt.QDoubleValidator(parent=self._qleAngleOffset)
        self._qleAngleOffset.setValidator(validator)
        widget.layout().addWidget(self._qleAngleOffset)
        self.LinkLineEditWithH5Variable(
            self._qleAngleOffset, "ANGLE_OFFSET_VALUE", float
        )

        # Since the OctaveH5 has to have the ANGLE_OFFSET parameter we are
        # defining it in regards of ANGLE_OFFSET_VALUE
        # this parameter is hidden and the user can't interacte with it
        self.linkGroupWithH5Variable(
            group=None,
            h5ParamName="ANGLE_OFFSET",
            setter=None,
            getter=self._getAngleOffsetParamVal,
        )

        return widget

    def _getAngleOffsetParamVal(self):
        """

        :return: True if ANGLE_OFFSET_VALUE != 0
        """
        return int(float(self._qleAngleOffset.text()) != 0.0)

    def _angleOffsetChanged(self):
        if self._isLoading is False:
            value = self._getAngleOffsetParamVal()
            self._recons_params._set_parameter_value(
                parameter="ANGLE_OFFSET", value=value
            )

    def _angleOffsetValueChanged(self):
        if self._isLoading is False:
            value = float(self._qleAngleOffset.text())
            self._recons_params._set_parameter_value(
                parameter="ANGLE_OFFSET_VALUE", value=value
            )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "26/10/2020"


from tomoscan.esrf.hdf5scan import ImageKey as _ImageKey
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from nxtomomill.utils import change_image_key_control as _change_image_key_control
from tomwer.core.process.task import Task
from tomoscan.esrf.hdf5scan import ImageKey
from tomwer.core.scan.scanfactory import ScanFactory
import nxtomomill.version
import logging

_logger = logging.getLogger(__name__)


IMAGE_KEYS = {
    "projection": _ImageKey.PROJECTION,
    "invalid": _ImageKey.INVALID,
    "dark": _ImageKey.DARK_FIELD,
    "flat": _ImageKey.FLAT_FIELD,
}


def change_image_key_control(scan: HDF5TomoScan, config: dict) -> TomwerScanBase:
    """

    :param scan:
    :param config:
    :raises KeyError: if 'frames_indexes' or 'image_key_control_value' are
                      not in config
    :return:
    """
    if scan is None:
        return
    elif not isinstance(scan, HDF5TomoScan):
        raise ValueError(
            "Image key control only handle HDF5TomoScan and "
            "not {}".format(type(scan))
        )

    if "modifications" not in config:
        raise KeyError("modifications are not provided")
    else:
        modifications = config["modifications"]
        if modifications is None:
            modifications = {}

    image_keys_set = set(modifications.values())
    image_keys_set = set(
        [ImageKey.from_value(image_key) for image_key in image_keys_set]
    )
    for image_key_type in image_keys_set:
        frame_indexes_dict = dict(
            filter(lambda item: item[1] is image_key_type, modifications.items())
        )
        frame_indexes = tuple(frame_indexes_dict.keys())
        _logger.warning("will modify {} to {}".format(frame_indexes, image_key_type))
        _change_image_key_control(
            file_path=scan.master_file,
            entry=scan.entry,
            frames_indexes=frame_indexes,
            image_key_control_value=image_key_type.value,
            logger=_logger,
        )
    scan.clear_caches()
    scan._frames = None
    return scan


class ImageKeyEditor(Task, input_names=("data",), output_names=("data",)):
    def __init__(self, varinfo=None, inputs=None, label=None):
        super().__init__(varinfo=varinfo, inputs=inputs, label=label)
        self.set_configuration(inputs.get("image_key_edition", {}))

    def run(self):
        scan = self.inputs.data
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                "scan is expected to be a dict or an instance "
                "of TomwerScanBase. Not {}".format(type(scan))
            )
        if not isinstance(scan, HDF5TomoScan):
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )

        change_image_key_control(scan=scan, config=self.get_configuration())
        config = self.get_configuration()
        modif_keys = list(config.get("modifications", {}).keys())
        new_modif = {}
        for key in modif_keys:
            value = config["modifications"][key]
            config["modifications"][str(key)] = value
        config["modifications"] = new_modif
        with scan.acquire_process_file_lock():
            self.register_process(
                process_file=scan.process_file,
                entry=scan.entry,
                configuration=config,
                results={},
                process_index=scan.pop_process_index(),
                overwrite=True,
            )
        self.outputs.data = scan

    @staticmethod
    def program_name():
        return "nxtomomill.utils.change_image_key_control"

    @staticmethod
    def program_version():
        return nxtomomill.version.version

    @staticmethod
    def definition():
        return "Modify image keys on an NXTomo"



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "03/05/2019"


from silx.gui import qt
from tomwer.core.process.reconstruction.ftseries.params import (
    ReconsParams,
    beamgeo,
    axis,
    paganin,
)
from tomwer.core.process.reconstruction.ftseries.params import pyhst, ft
from tomwer.core.process.reconstruction.darkref import params as dkrf

import logging

logger = logging.getLogger(__name__)


class _QAxisRP(axis.AxisRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        axis.AxisRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class _QBeamGeoRP(beamgeo.BeamGeoRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        beamgeo.BeamGeoRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class _QDKRFRP(dkrf.DKRFRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        dkrf.DKRFRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class _QFTRP(ft.FTRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        ft.FTRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class _QPaganinRP(paganin.PaganinRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        paganin.PaganinRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class _QPyhstRP(pyhst.PyhstRP, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        pyhst.PyhstRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()


class QReconsParams(ReconsParams, qt.QObject):

    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self, empty=False):
        ReconsParams.__init__(self, empty=empty)
        qt.QObject.__init__(self)

        # connect signal / Slot
        if self._ft:
            self._ft.sigChanged.connect(self.changed)
        if self._pyhst:
            self._pyhst.sigChanged.connect(self.changed)
        if self.axis:
            self._axis.sigChanged.connect(self.changed)
        if self.paganin:
            self._paganin.sigChanged.connect(self.changed)
        if self.beam_geo:
            self._beam_geo.sigChanged.connect(self.changed)
        if self.dkrf:
            self._dkrf.sigChanged.connect(self.changed)

    def _createSubParamsSet(self, empty):
        self._ft = None if empty else _QFTRP()
        self._pyhst = None if empty else _QPyhstRP()
        self._axis = None if empty else _QAxisRP()
        self._paganin = None if empty else _QPaganinRP()
        self._beam_geo = None if empty else _QBeamGeoRP()
        self._dkrf = None if empty else _QDKRFRP()

    def changed(self):
        self.sigChanged.emit()

    def _copy_ft_rp(self, other_rp):
        create_connection = self.ft is None
        ReconsParams._copy_ft_rp(self, other_rp=other_rp)
        if create_connection:
            self._ft.sigChanged.connect(self.changed)

    def _copy_pyhst_rp(self, other_rp):
        create_connection = self.pyhst is None
        ReconsParams._copy_pyhst_rp(self, other_rp=other_rp)
        if create_connection:
            self._pyhst.sigChanged.connect(self.changed)

    def _copy_axis_rp(self, other_rp):
        create_connection = self.axis is None
        ReconsParams._copy_axis_rp(self, other_rp=other_rp)
        if create_connection:
            self._axis.sigChanged.connect(self.changed)

    def _copy_paganin_rp(self, other_rp):
        create_connection = self.paganin is None
        ReconsParams._copy_paganin_rp(self, other_rp=other_rp)
        if create_connection:
            self._paganin.sigChanged.connect(self.changed)

    def _copy_beam_geo_rp(self, other_rp):
        create_connection = self.beam_geo is None
        ReconsParams._copy_beam_geo_rp(self, other_rp=other_rp)
        if create_connection:
            self._beam_geo.sigChanged.connect(self.changed)

    def _copy_dkrf_rp(self, other_rp):
        create_connection = self.dkrf is None
        ReconsParams._copy_dkrf_rp(self, other_rp=other_rp)
        if create_connection:
            self._dkrf.sigChanged.connect(self.changed)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains utils for inputs and outputs
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/04/2021"


import contextlib
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
import h5py


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError("url should be an instance of DataUrl")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(filename=self._url.file_path(), mode="r")
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(filename=self._url.file_path(), mode="r")
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError("Data path should point to a dtaset (h5py.Dataset)")
        return entry
